"""Test the Store interface

Many methods simply proxy to config/metadata/data store implementations, but there is some
cross-coordination and there are some convenience methods implemented at this layer.
"""
import os

import numpy as np
import numpy.testing
from pytest import fixture, raises
from smif.data_layer import Store
from smif.data_layer.data_array import DataArray
from smif.data_layer.memory_interface import (MemoryConfigStore,
                                              MemoryDataStore,
                                              MemoryMetadataStore)
from smif.exception import SmifDataError, SmifDataNotFoundError
from smif.metadata import Spec


@fixture
def store():
    """Store fixture
    """
    # implement each part using the memory classes, simpler than mocking
    # each other implementation of a part is tested fully by e.g. test_config_store.py
    return Store(
        config_store=MemoryConfigStore(),
        metadata_store=MemoryMetadataStore(),
        data_store=MemoryDataStore()
    )


@fixture
def full_store(store, get_sos_model, get_sector_model, energy_supply_sector_model,
               sample_scenarios, model_run, sample_dimensions):
    for dim in sample_dimensions:
        store.write_dimension(dim)
    store.write_model(get_sector_model)
    store.write_model(energy_supply_sector_model)
    for scenario in sample_scenarios:
        store.write_scenario(scenario)
    store.write_sos_model(get_sos_model)
    store.write_model_run(model_run)
    return store


class TestStoreConfig():
    def test_model_runs(self, store, minimal_model_run):
        # write
        store.write_model_run(minimal_model_run)
        # read all
        assert store.read_model_runs() == [minimal_model_run]
        # read one
        assert store.read_model_run(minimal_model_run['name']) == minimal_model_run
        # update
        store.update_model_run(minimal_model_run['name'], minimal_model_run)
        # delete
        store.delete_model_run(minimal_model_run['name'])
        assert store.read_model_runs() == []

    def test_read_model_run_sorted(self, store, minimal_model_run):
        y_model_run = {'name': 'y'}
        z_model_run = {'name': 'z'}

        store.write_model_run(minimal_model_run)
        store.write_model_run(z_model_run)
        store.write_model_run(y_model_run)

        expected = [minimal_model_run, y_model_run, z_model_run]
        assert store.read_model_runs() == expected

    def test_sos_models(self, store, get_sos_model, get_sector_model,
                        energy_supply_sector_model, sample_scenarios):
        # write
        store.write_model(get_sector_model)
        store.write_model(energy_supply_sector_model)
        for scenario in sample_scenarios:
            store.write_scenario(scenario)
        store.write_sos_model(get_sos_model)
        # read all
        assert store.read_sos_models() == [get_sos_model]
        # read one
        assert store.read_sos_model(get_sos_model['name']) == get_sos_model
        # update
        store.update_sos_model(get_sos_model['name'], get_sos_model)
        # delete
        store.delete_sos_model(get_sos_model['name'])
        assert store.read_sos_models() == []

    def test_models(self, store, get_sector_model, sample_dimensions):
        # setup
        for dim in sample_dimensions:
            store.write_dimension(dim)
        # write
        store.write_model(get_sector_model)
        # read all
        assert store.read_models() == [get_sector_model]
        # read one
        assert store.read_model(get_sector_model['name']) == get_sector_model
        # update
        store.update_model(get_sector_model['name'], get_sector_model)
        # delete
        store.delete_model(get_sector_model['name'])
        assert store.read_models() == []
        # teardown
        for dim in sample_dimensions:
            store.delete_dimension(dim['name'])

    def test_models_skip_coords(self, store, get_sector_model, get_sector_model_no_coords):
        # write
        store.write_model(get_sector_model)
        # read all
        assert store.read_models(skip_coords=True) == [get_sector_model_no_coords]
        # read one
        assert store.read_model(get_sector_model['name'],
                                skip_coords=True) == get_sector_model_no_coords
        # update
        store.update_model(get_sector_model['name'], get_sector_model)
        # delete
        store.delete_model(get_sector_model['name'])
        assert store.read_models() == []

    def test_scenarios(self, store, scenario, sample_dimensions):
        # setup
        for dim in sample_dimensions:
            store.write_dimension(dim)
        # write
        store.write_scenario(scenario)
        # read all
        assert store.read_scenarios() == [scenario]
        # read one
        assert store.read_scenario(scenario['name']) == scenario
        # update
        store.update_scenario(scenario['name'], scenario)
        # delete
        store.delete_scenario(scenario['name'])
        assert store.read_scenarios() == []
        # teardown
        for dim in sample_dimensions:
            store.delete_dimension(dim['name'])

    def test_scenarios_skip_coords(self, store, scenario, scenario_no_coords):
        # write
        store.write_scenario(scenario)
        # read all
        assert store.read_scenarios(skip_coords=True) == [scenario_no_coords]
        # read one
        assert store.read_scenario(scenario['name'], skip_coords=True) == scenario_no_coords
        # update
        store.update_scenario(scenario['name'], scenario)
        # delete
        store.delete_scenario(scenario['name'])
        assert store.read_scenarios() == []

    def test_scenario_variants(self, store, scenario):
        scenario_name = scenario['name']
        old_variant = scenario['variants'][0]
        # write
        store.write_scenario(scenario)
        new_variant = {
            'name': 'high',
            'description': 'Mortality (High)',
            'data': {
                'mortality': 'mortality_high.csv'
            }
        }
        store.write_scenario_variant(scenario_name, new_variant)
        # read all
        both = store.read_scenario_variants(scenario_name)
        assert both == [old_variant, new_variant] or both == [new_variant, old_variant]
        # read one
        assert store.read_scenario_variant(scenario_name, old_variant['name']) == old_variant
        # update
        store.update_scenario_variant(scenario_name, old_variant['name'], old_variant)
        # delete
        store.delete_scenario_variant(scenario_name, old_variant['name'])
        assert store.read_scenario_variants(scenario_name) == [new_variant]

    def test_prepare_scenario(self, store, scenario, scenario_2_variants,
                              scenario_no_variant, sample_dimensions):
        for dim in sample_dimensions:
            store.write_dimension(dim)
        # Insert template_scenario dict in underlying
        # MemoryConfigStore
        store.write_scenario(scenario)
        store.write_scenario(scenario_2_variants)
        store.write_scenario(scenario_no_variant)

        list_of_variants = range(1, 4)

        # Must raise exception if scenario defines > 1 variants
        with raises(SmifDataError) as ex:
            store.prepare_scenario(scenario_2_variants['name'], list_of_variants)
        assert "must define one unique template variant" in str(ex.value)

        # Must raise exception if scenario defines 0 variants
        with raises(SmifDataError) as ex:
            store.prepare_scenario(scenario_no_variant['name'], list_of_variants)
        assert "must define one unique template variant" in str(ex.value)

        store.prepare_scenario(scenario['name'], list_of_variants)

        updated_scenario = store.read_scenario(scenario['name'])

        assert len(updated_scenario['variants']) == 3
        assert updated_scenario['variants'][0]['name'] == 'mortality_001'
        new_variant = store.read_scenario_variant(scenario['name'], 'mortality_001')
        #
        assert new_variant['name'] == 'mortality_001'
        assert new_variant['description'] == 'mortality variant number 001'
        assert new_variant['data']['mortality'] == 'mortality_low001.csv'

        assert updated_scenario['variants'][1]['name'] == 'mortality_002'
        new_variant = store.read_scenario_variant(scenario['name'], 'mortality_002')
        assert new_variant['name'] == 'mortality_002'
        assert new_variant['description'] == 'mortality variant number 002'
        assert new_variant['data']['mortality'] == 'mortality_low002.csv'

        assert updated_scenario['variants'][2]['name'] == 'mortality_003'
        new_variant = store.read_scenario_variant(scenario['name'], 'mortality_003')
        assert new_variant['name'] == 'mortality_003'
        assert new_variant['description'] == 'mortality variant number 003'
        assert new_variant['data']['mortality'] == 'mortality_low003.csv'

    def test_prepare_model_runs(self, store, model_run, sample_scenarios, sample_dimensions):
        for dim in sample_dimensions:
            store.write_dimension(dim)
        scenario = sample_scenarios[0]
        store.write_model_run(model_run)
        store.write_strategies(model_run['name'], model_run['strategies'])
        store.write_scenario(scenario)

        # Generate 2 model runs for variants Low and High
        store.prepare_model_runs(model_run['name'], scenario['name'], 0, 1)
        list_of_mr = store.read_model_runs()
        assert len(list_of_mr) == 3
        assert list_of_mr[0] == model_run
        assert list_of_mr[1]['name'] == model_run['name'] + '_' + scenario['variants'][0][
            'name']
        assert list_of_mr[2]['name'] == model_run['name'] + '_' + scenario['variants'][1][
            'name']
        store.delete_model_run(list_of_mr[1]['name'])
        store.delete_model_run(list_of_mr[2]['name'])
        # Generate only one model run for variant Low
        store.prepare_model_runs(model_run['name'], scenario['name'], 0, 0)
        list_of_mr = store.read_model_runs()
        assert len(list_of_mr) == 2
        assert list_of_mr[0] == model_run
        assert list_of_mr[1]['name'] == model_run['name'] + '_' + scenario['variants'][0][
            'name']

        # Tidy up batch file
        os.remove('{}.batch'.format(model_run['name']))

    def test_narratives(self, store, get_sos_model):
        store.write_sos_model(get_sos_model)
        expected = get_sos_model['narratives'][0]
        # read one
        assert store.read_narrative(get_sos_model['name'], expected['name']) == expected

    def test_strategies(self, store, strategies):
        model_run_name = 'test_modelrun'
        # write
        store.write_strategies(model_run_name, strategies)
        # read
        assert store.read_strategies(model_run_name) == strategies


class TestStoreMetadata():
    def test_units(self, store, unit_definitions):
        # write
        store.write_unit_definitions(unit_definitions)
        # read
        assert store.read_unit_definitions() == unit_definitions

    def test_dimensions(self, store, dimension):
        # write
        store.write_dimension(dimension)
        # read all
        assert store.read_dimensions() == [dimension]
        # read one
        assert store.read_dimension(dimension['name']) == dimension
        # update
        store.update_dimension(dimension['name'], dimension)
        # delete
        store.delete_dimension(dimension['name'])
        assert store.read_dimensions() == []


class TestStoreData():
    @fixture(scope='function')
    def setup(self, store, sample_dimensions, scenario,
              sample_scenario_data):
        for dim in sample_dimensions:
            store.write_dimension(dim)
        store.write_scenario(scenario)
        # pick out single sample
        key = next(iter(sample_scenario_data))
        scenario_name, variant_name, variable = key
        scenario_variant_data = sample_scenario_data[key]
        # write
        store.write_scenario_variant_data(
            scenario_name, variant_name, scenario_variant_data
        )
        return key, scenario_variant_data

    def test_convert_strategies_data(self, empty_store, store, strategies):
        src_store = store
        tgt_store = empty_store  # Store with target data format
        model_run_name = 'test_modelrun'
        # write
        src_store.write_strategies(model_run_name, strategies)
        # convert
        src_store.convert_strategies_data(model_run_name, tgt_store)
        # assert
        for strategy in strategies:
            if 'interventions' in strategy:  # If the stategy fixture defines interventions
                expected = src_store.read_strategy_interventions(strategy)
                assert expected == tgt_store.read_strategy_interventions(strategy)

    def test_scenario_variant_data(self, store,
                                   setup):
        # The sample_scenario_data fixture provides data with a spec including timestep
        # dimension containing a single coordinate of 2015. Note the asymmetry in the write
        # and read methods here: writing requires the full DataArray object with the full
        # spec including timestep, but the reading requires a specific timestep to be supplied.
        # The data read back in, therefore, has lower dimensionality.

        key, scenario_variant_data = setup
        scenario_name, variant_name, variable = key

        assert store.read_scenario_variant_data(scenario_name, variant_name, variable,
                                                2015, assert_exists=True)
        # Read 2015
        actual = store.read_scenario_variant_data(
            scenario_name, variant_name, variable, 2015
        )
        assert (actual.data == scenario_variant_data.data[0]).all()

        # Read 2016
        actual = store.read_scenario_variant_data(
            scenario_name, variant_name, variable, 2016
        )
        assert (actual.data == scenario_variant_data.data[1]).all()

    def test_convert_scenario_data(self, empty_store, store, sample_dimensions, scenario,
                                   sample_scenario_data, model_run):
        src_store = store
        tgt_store = empty_store  # Store with target data format
        # setup
        model_run['scenarios'] = {'mortality': 'low'}
        model_run['timesteps'] = [2015, 2016]
        src_store.write_model_run(model_run)
        tgt_store.write_model_run(model_run)
        for dim in sample_dimensions:
            src_store.write_dimension(dim)
            tgt_store.write_dimension(dim)

        src_store.write_scenario(scenario)
        tgt_store.write_scenario(scenario)
        # pick out single sample
        key = next(iter(sample_scenario_data))
        scenario_name, variant_name, variable = key
        scenario_variant_data = sample_scenario_data[key]
        # write
        src_store.write_scenario_variant_data(
            scenario_name, variant_name, scenario_variant_data
        )
        src_store.convert_scenario_data(model_run['name'], tgt_store)

        for variant in src_store.read_scenario_variants(scenario_name):
            for variable in variant['data']:
                expected = src_store.read_scenario_variant_data(
                    scenario_name, variant['name'], variable, timesteps=model_run['timesteps'])
                result = tgt_store.read_scenario_variant_data(
                    scenario_name, variant['name'], variable, timesteps=model_run['timesteps'])
                assert result == expected

    def test_scenario_variant_data_mult_one_year(self, store, setup):
        key, scenario_variant_data = setup
        scenario_name, variant_name, variable = key

        actual = store.read_scenario_variant_data(
            scenario_name, variant_name, variable, timesteps=[2016])

        assert (actual.data == [scenario_variant_data.data[1]]).all()

    def test_scenario_variant_data_mult_mult_years(self, store, setup):

        key, scenario_variant_data = setup
        scenario_name, variant_name, variable = key

        actual = store.read_scenario_variant_data(
            scenario_name, variant_name, variable, timesteps=[2015, 2016])

        assert (actual.data == scenario_variant_data.data).all()

    def test_narrative_variant_data(self, store, sample_dimensions, get_sos_model,
                                    get_sector_model, energy_supply_sector_model,
                                    sample_narrative_data):
        # Setup
        for dim in sample_dimensions:
            store.write_dimension(dim)
        store.write_sos_model(get_sos_model)
        store.write_model(get_sector_model)
        store.write_model(energy_supply_sector_model)
        # pick out single sample
        key = (
            'energy',
            'technology',
            'high_tech_dsm',
            'smart_meter_savings'
        )
        sos_model_name, narrative_name, variant_name, param_name = key
        narrative_variant_data = sample_narrative_data[key]
        # write
        store.write_narrative_variant_data(
            sos_model_name, narrative_name, variant_name, narrative_variant_data)

        assert store.read_narrative_variant_data(
            sos_model_name, narrative_name, variant_name, param_name, assert_exists=True)
        # read
        actual = store.read_narrative_variant_data(
            sos_model_name, narrative_name, variant_name, param_name)
        assert actual == narrative_variant_data

    def test_convert_narrative_data(self, empty_store, store, sample_dimensions, get_sos_model,
                                    get_sector_model, energy_supply_sector_model,
                                    sample_narrative_data, model_run):
        src_store = store
        tgt_store = empty_store
        # Setup
        for dim in sample_dimensions:
            src_store.write_dimension(dim)
            tgt_store.write_dimension(dim)
        src_store.write_sos_model(get_sos_model)
        src_store.write_model(get_sector_model)
        src_store.write_model(energy_supply_sector_model)
        tgt_store.write_sos_model(get_sos_model)
        tgt_store.write_model(get_sector_model)
        tgt_store.write_model(energy_supply_sector_model)

        for narrative in get_sos_model['narratives']:
            for variant in narrative['variants']:
                for param_name in variant['data']:
                    key = (
                        get_sos_model['name'],
                        narrative['name'],
                        variant['name'],
                        param_name
                    )
                    narrative_variant_data = sample_narrative_data[key]
                    # write
                    src_store.write_narrative_variant_data(
                        get_sos_model['name'], narrative['name'], variant['name'],
                        narrative_variant_data)

        src_store.convert_narrative_data(model_run['sos_model'], tgt_store)

        for narrative in get_sos_model['narratives']:
            for variant in narrative['variants']:
                for param_name in variant['data']:
                    expected = src_store.read_narrative_variant_data(
                        get_sos_model['name'], narrative['name'], variant['name'], param_name)
                    result = tgt_store.read_narrative_variant_data(
                        get_sos_model['name'], narrative['name'], variant['name'], param_name)
                    assert result == expected

    def test_model_parameter_default(self, store, get_multidimensional_param,
                                     get_sector_model, sample_dimensions):
        param_data = get_multidimensional_param
        for dim in sample_dimensions:
            store.write_dimension(dim)
        for dim in param_data.dims:
            store.write_dimension({'name': dim, 'elements': param_data.dim_elements(dim)})
        get_sector_model['parameters'] = [param_data.spec.as_dict()]
        store.write_model(get_sector_model)
        # write
        store.write_model_parameter_default(
            get_sector_model['name'], param_data.name, param_data)
        assert store.read_model_parameter_default(get_sector_model['name'],
                                                  param_data.name, assert_exists=True)
        # read
        actual = store.read_model_parameter_default(get_sector_model['name'], param_data.name)
        assert actual == param_data

    def test_convert_model_parameter_default_data(self, empty_store, store,
                                                  get_multidimensional_param,
                                                  get_sector_model, sample_dimensions):
        src_store = store
        tgt_store = empty_store
        param_data = get_multidimensional_param
        for store in [src_store, tgt_store]:
            for dim in sample_dimensions:
                store.write_dimension(dim)
            for dim in param_data.dims:
                store.write_dimension({'name': dim, 'elements': param_data.dim_elements(dim)})

            get_sector_model['parameters'] = [param_data.spec.as_dict()]
            store.write_model(get_sector_model)

        # write
        src_store.write_model_parameter_default(
            get_sector_model['name'], param_data.name, param_data)
        # convert
        src_store.convert_model_parameter_default_data(get_sector_model['name'], tgt_store)

        expected = src_store.read_model_parameter_default(
            get_sector_model['name'], param_data.name)
        result = tgt_store.read_model_parameter_default(
            get_sector_model['name'], param_data.name)

        assert result == expected

    def test_interventions(self, store, sample_dimensions, get_sector_model, interventions):
        # setup
        for dim in sample_dimensions:
            store.write_dimension(dim)
        store.write_model(get_sector_model)
        # write
        store.write_interventions(get_sector_model['name'], interventions)
        # read
        assert store.read_interventions(get_sector_model['name']) == interventions

    def test_convert_interventions_data(self, empty_store, store, sample_dimensions,
                                        get_sector_model, interventions):
        src_store = store
        tgt_store = empty_store
        get_sector_model['interventions'] = ['energy_demand.csv']
        # setup
        for dim in sample_dimensions:
            src_store.write_dimension(dim)
            tgt_store.write_dimension(dim)
        src_store.write_model(get_sector_model)
        tgt_store.write_model(get_sector_model)
        # write
        src_store.write_interventions(get_sector_model['name'], interventions)

        src_store.convert_interventions_data(get_sector_model['name'], tgt_store)

        assert interventions == tgt_store.read_interventions(get_sector_model['name'])

    def test_read_write_interventions_file(self, store, sample_dimensions,
                                           get_sector_model, interventions):
        # setup
        for dim in sample_dimensions:
            store.write_dimension(dim)
        get_sector_model['interventions'] = ['path']
        store.write_model(get_sector_model)
        # write
        store.write_interventions_file(get_sector_model['name'], 'path', interventions)
        # check data existence
        assert store.read_interventions_file(
            get_sector_model['name'], 'path', assert_exists=True)

        result = store.read_interventions_file(get_sector_model['name'], 'path')
        assert result == interventions

    def test_initial_conditions(self, store, sample_dimensions, initial_conditions,
                                get_sos_model, get_sector_model, energy_supply_sector_model,
                                minimal_model_run):
        # setup
        for dim in sample_dimensions:
            store.write_dimension(dim)
        store.write_sos_model(get_sos_model)
        store.write_model_run(minimal_model_run)
        store.write_model(get_sector_model)
        store.write_model(energy_supply_sector_model)
        # write
        store.write_initial_conditions(get_sector_model['name'], initial_conditions)
        # read
        assert store.read_initial_conditions(get_sector_model['name']) == initial_conditions
        # read all for a model run
        actual = store.read_all_initial_conditions(minimal_model_run['name'])
        assert actual == initial_conditions

    def test_convert_initial_conditions_data(self, empty_store, store, sample_dimensions,
                                             initial_conditions, get_sos_model,
                                             get_sector_model, energy_supply_sector_model,
                                             minimal_model_run):
        src_store = store
        tgt_store = empty_store
        get_sector_model['initial_conditions'] = ['energy_demand.csv']
        for store in [src_store, tgt_store]:
            for dim in sample_dimensions:
                store.write_dimension(dim)
            store.write_sos_model(get_sos_model)
            store.write_model_run(minimal_model_run)
            store.write_model(get_sector_model)
            store.write_model(energy_supply_sector_model)

        src_store.write_initial_conditions(get_sector_model['name'], initial_conditions)

        src_store.convert_initial_conditions_data(get_sector_model['name'], tgt_store)

        assert initial_conditions == tgt_store.read_initial_conditions(
            get_sector_model['name'])

    def test_read_write_initial_conditions_file(self, store, sample_dimensions,
                                                get_sector_model, initial_conditions):
        # setup
        for dim in sample_dimensions:
            store.write_dimension(dim)
        get_sector_model['initial_conditions'] = ['path']
        store.write_model(get_sector_model)
        # write
        store.write_initial_conditions_file(get_sector_model['name'],
                                            'path', initial_conditions)
        assert store.read_initial_conditions_file(get_sector_model['name'], 'path',
                                                  assert_exists=True)
        result = store.read_initial_conditions_file(get_sector_model['name'], 'path')

        assert result == initial_conditions

    def test_state(self, store, state):
        # write
        store.write_state(state, 'model_run_name', 0, 0)
        # read
        assert store.read_state('model_run_name', 0, 0) == state

    def test_conversion_coefficients(self, store, conversion_coefficients):
        # write
        store.write_coefficients(
            'source_dim', 'sink_dim', conversion_coefficients)
        # read
        numpy.testing.assert_equal(
            store.read_coefficients('source_dim', 'sink_dim'),
            conversion_coefficients
        )

    def test_results(self, store, sample_results):
        # write
        store.write_results(sample_results, 'model_run_name', 'model_name', 0)
        # read
        spec = sample_results.spec
        assert store.read_results('model_run_name', 'model_name', spec, 0) == sample_results
        # check
        assert store.available_results('model_run_name') == [
            (0, None, 'model_name', spec.name)
        ]
        # delete
        store.clear_results('model_run_name')
        assert not store.available_results('model_run_name')

    def test_no_completed_jobs(self, full_store):
        expected = []
        actual = full_store.completed_jobs('unique_model_run_name')
        assert actual == expected

    def test_expected_model_outputs(self, full_store):
        actual = full_store.expected_model_outputs('unique_model_run_name')
        expected = [('energy_demand', 'gas_demand')]
        assert actual == expected

    def test_some_completed_jobs(self, full_store, sample_gas_demand_results):
        expected = [
            (2015, 0, 'energy_demand'),
            (2020, 0, 'energy_demand'),
        ]
        full_store.write_results(
            sample_gas_demand_results, 'unique_model_run_name', 'energy_demand', 2015, 0)
        full_store.write_results(
            sample_gas_demand_results, 'unique_model_run_name', 'energy_demand', 2020, 0)
        actual = full_store.completed_jobs('unique_model_run_name')
        assert actual == expected

    def test_filter_complete_available_results(self, store):
        available_results = [
            (2020, 0, 'test_model', 'output_a'),
            (2020, 0, 'test_model', 'output_b'),
            (2025, 0, 'test_model', 'output_a'),
            (2030, 0, 'other_model', 'output_other'),
        ]
        model_outputs = [
            ('test_model', 'output_a'),
            ('test_model', 'output_b'),
        ]
        expected = [
            (2020, 0, 'test_model')
        ]
        actual = store.filter_complete_available_results(available_results, model_outputs)
        assert actual == expected

    def test_warm_start(self, store, sample_results):
        assert store.prepare_warm_start('test_model_run') is None
        timestep = 2020
        store.write_results(sample_results, 'test_model_run', 'model_name', timestep)
        assert store.prepare_warm_start('test_model_run') == timestep

    def test_canonical_available_results(self, store, sample_results):

        store.write_results(sample_results, 'model_run_name', 'model_name', 2010, 0)
        store.write_results(sample_results, 'model_run_name', 'model_name', 2015, 0)
        store.write_results(sample_results, 'model_run_name', 'model_name', 2010, 1)
        store.write_results(sample_results, 'model_run_name', 'model_name', 2015, 1)
        store.write_results(sample_results, 'model_run_name', 'model_name', 2020, 1)

        output_name = sample_results.spec.name

        correct_results = set()
        correct_results.add((2010, 0, 'model_name', output_name))
        correct_results.add((2015, 0, 'model_name', output_name))
        correct_results.add((2020, 0, 'model_name', output_name))

        assert (store.canonical_available_results('model_run_name') == correct_results)

    def test_canonical_expected_results(
            self, store, sample_dimensions, get_sos_model, get_sector_model,
            energy_supply_sector_model, model_run
    ):

        for dim in sample_dimensions:
            store.write_dimension(dim)
        store.write_sos_model(get_sos_model)
        store.write_model_run(model_run)
        store.write_model(get_sector_model)
        store.write_model(energy_supply_sector_model)

        correct_results = set()
        correct_results.add((2015, 0, 'energy_demand', 'gas_demand'))
        correct_results.add((2020, 0, 'energy_demand', 'gas_demand'))
        correct_results.add((2025, 0, 'energy_demand', 'gas_demand'))

        assert (store.canonical_expected_results(model_run['name']) == correct_results)

    def test_canonical_missing_results(
            self, store, sample_dimensions, get_sos_model, get_sector_model,
            energy_supply_sector_model, model_run
    ):

        for dim in sample_dimensions:
            store.write_dimension(dim)
        store.write_sos_model(get_sos_model)
        store.write_model_run(model_run)
        store.write_model(get_sector_model)
        store.write_model(energy_supply_sector_model)

        # All the results are missing
        missing_results = set()
        missing_results.add((2015, 0, 'energy_demand', 'gas_demand'))
        missing_results.add((2020, 0, 'energy_demand', 'gas_demand'))
        missing_results.add((2025, 0, 'energy_demand', 'gas_demand'))

        assert (store.canonical_missing_results(model_run['name']) == missing_results)

        spec = Spec(name='gas_demand', dtype='float')
        data = np.array(1, dtype=float)
        fake_data = DataArray(spec, data)

        store.write_results(fake_data, model_run['name'], 'energy_demand', 2015, 0)
        missing_results.remove((2015, 0, 'energy_demand', 'gas_demand'))

        assert (store.canonical_missing_results(model_run['name']) == missing_results)

    def test_get_results(self):
        # This is difficult to test without fixtures defining an entire canonical project.
        # See smif issue #304 (https://github.com/nismod/smif/issues/304).
        # Todo: mock a store with known results that can be obtained with get_results(...)
        # This requires a model run with sector model, and a sector model with valid inputs and
        # outputs, and results with valid spec, etc. Some of this functionality exists in
        # fixtures provided in `conftest.py`.
        pass


class TestWrongRaises:

    def test_narrative_variant(self, store, sample_dimensions,
                               get_sos_model, get_sector_model,
                               energy_supply_sector_model,
                               sample_narrative_data):
        # Setup
        for dim in sample_dimensions:
            store.write_dimension(dim)
        store.write_sos_model(get_sos_model)
        store.write_model(get_sector_model)
        store.write_model(energy_supply_sector_model)
        # pick out single sample
        key = (
            'energy',
            'technology',
            'high_tech_dsm',
            'smart_meter_savings'
        )
        sos_model_name, narrative_name, variant_name, param_name = key

        with raises(SmifDataNotFoundError) as ex:
            store.read_narrative_variant_data(
                sos_model_name, narrative_name, 'bla', param_name)

        expected = "Variant name 'bla' does not exist in narrative 'technology'"

        assert expected in str(ex.value)

    def test_narrative_name(self, store, sample_dimensions, get_sos_model,
                            get_sector_model, energy_supply_sector_model,
                            sample_narrative_data):
        # Setup
        for dim in sample_dimensions:
            store.write_dimension(dim)
        store.write_sos_model(get_sos_model)
        store.write_model(get_sector_model)
        store.write_model(energy_supply_sector_model)
        # pick out single sample
        key = (
            'energy',
            'technology',
            'high_tech_dsm',
            'smart_meter_savings'
        )
        sos_model_name, narrative_name, variant_name, param_name = key

        with raises(SmifDataNotFoundError) as ex:
            store.read_narrative_variant_data(
                sos_model_name, 'bla', variant_name, param_name)

        expected = "Narrative name 'bla' does not exist in sos_model 'energy'"
        assert expected in str(ex.value)



"""Schedulers are used to run models.

The defaults provided allow model runs to be scheduled as subprocesses,
or individual models to be called in series.

Future implementations may interface with common schedulers to queue
up models to run in parallel and/or distributed.

Calls smif run ... on the selected model in a sub process, where ...
are the options set in the app for info/debug messages, warm start
and output format.
"""
import subprocess
from collections import defaultdict
from datetime import datetime


class SubProcessRunScheduler(object):
    """The scheduler can run instances of smif as a subprocess
    and can provide information whether the modelrun is running,
    is done or has failed.
    """
    def __init__(self):
        self._status = defaultdict(lambda: 'unstarted')
        self._process = {}
        self._output = defaultdict(str)
        self._err = {}
        self.lock = False

    def add(self, model_run_name, args):
        """Add a model_run to the Modelrun scheduler.

        Parameters
        ----------
        model_run_name: str
            Name of the modelrun
        args: dict
            Arguments for the command-line interface

        Exception
        ---------
        Exception
            When the modelrun was already started

        Notes
        -----
        There is no queuing mechanism implemented, each `add`
        will directly start a subprocess. This means that it
        is possible to run multiple modelruns concurrently.
        This may cause conflicts, it depends on the
        implementation whether a certain sector model / wrapper
        touches the filesystem or other shared resources.
        """
        if self._status[model_run_name] != 'running':
            self._output[model_run_name] = ''
            self._status[model_run_name] = 'queing'

            smif_call = (
                'smif run ' +
                '-'*(int(args['verbosity']) > 0) + 'v'*int(args['verbosity']) + ' ' +
                model_run_name + ' ' +
                '-d' + ' ' + args['directory'] + ' ' +
                '-w'*args['warm_start'] + ' '*args['warm_start'] +
                '-i' + ' ' + args['output_format']
            )

            self._process[model_run_name] = subprocess.Popen(
                smif_call,
                shell=True,
                stdout=subprocess.PIPE, stderr=subprocess.STDOUT
            )
            format_args = {
                'model_run_name': model_run_name,
                'datetime': datetime.now().strftime('%Y-%m-%d %H:%M:%S'),
                'pid': str(self._process[model_run_name].pid),
                'smif_call': smif_call,
                'colour': "\x1b[1;34m",
                'reset': "\x1b[0m",
                'space': " \x1b"
            }
            format_str = """\
                {colour}Modelrun{reset} {model_run_name}
                {colour}Time{reset}     {datetime}
                {colour}PID{reset}      {pid}
                {colour}Command{reset}  {smif_call}
                """
            format_str.replace(" ", "{space}")
            output = format_str.format(**format_args)
            output += "-" * 100 + "\n"
            self._output[model_run_name] = output
            self._status[model_run_name] = 'running'
        else:
            raise Exception('Model is already running.')

    def kill(self, model_run_name):
        """ Kill a Modelrun that is already running

        Parameters
        ----------
        model_run_name: str
            Name of the modelrun
        """
        if self._status[model_run_name] == 'running':
            self._process[model_run_name].kill()
            self._status[model_run_name] = 'stopped'

    def get_scheduler_type(self):
        return "default"

    def get_status(self, model_run_name):
        """Get the status from the Modelrun scheduler.

        Parameters
        ----------
        model_run_name: str
            Name of the modelrun

        Returns
        -------
        dict: A message containing the status, command-line
        output and error that can be directly sent back over
        the http api.

        Notes
        -----
        Possible status:

        unstarted:
            Model run was not started
        queing:
            Model run is waiting to be executed
        running:
            Model run is running
        stopped:
            Model run was stopped (killed) by user
        done:
            Model run was completed succesfully
        failed:
            Model run completed running with an exit code
        """
        if self._status[model_run_name] == 'running':
            if self.lock is False:
                self.lock = True
                for line in iter(self._process[model_run_name].stdout.readline, b''):
                    self._output[model_run_name] += line.decode()
                    self._process[model_run_name].stdout.flush()
                self.lock = False

            if self._process[model_run_name].poll() == 0:
                self._status[model_run_name] = 'done'
            elif self._process[model_run_name].poll() == 1:
                self._status[model_run_name] = 'failed'

        return {
            'status': self._status[model_run_name],
            'output': self._output[model_run_name]
        }



"""
Provides all the job run schedulers.
"""

# import classes for access like ::
#         from smif.controller.job import SerialJobScheduler
from smif.controller.job.serial_job_scheduler import SerialJobScheduler

# Define what should be imported as * ::
#         from smif.controller.job import *
__all__ = ['SerialJobScheduler']



import datetime
import logging
import logging.config
from collections import OrderedDict


# Make profiling methods available through the logger
def profiling_start(self, operation, key):
    time_placeholder = datetime.time(0, 0)
    level = sum(log[1]['stop'] == time_placeholder for log in logging.Logger._profile.items())
    logging.Logger._profile[(operation, key)] = {
        'start': datetime.datetime.now(),
        'stop': time_placeholder,
        'level': level
    }


def profiling_stop(self, operation, key):
    logging.Logger._profile[(operation, key)]['stop'] = datetime.datetime.now()


def summary(self, *args, **kws):

    if self.isEnabledFor(logging.INFO):
        summary = []
        columns = [30, 80, 12]
        column = "{:" + str(columns[0]) + "s}" + \
                 "{:" + str(columns[1]) + "s}" + \
                 "{:" + str(columns[2]) + "s}"

        level_width = max([profile['level'] for profile
                          in logging.Logger._profile.values()]) * 2
        total_width = sum(columns) + level_width

        # header
        summary.append(("{:*^" + str(total_width) + "s}").format(" Modelrun time profile "))
        summary.append(column.format('Function', 'Operation', 'Duration [hh:mm:ss]'))
        summary.append("*"*total_width)

        # body
        for profile in logging.Logger._profile.keys():

            # calculate time diff
            profile_data = logging.Logger._profile[profile]
            diff = profile_data['stop'] - profile_data['start']
            s = diff.total_seconds()
            time_spent = '{:02d}:{:02d}:{:02d}'.format(
                int(s // 3600), int(s % 3600 // 60), int(s % 60))

            # trunctuate long lines
            if len(profile[0]) > columns[0]-2:
                func = profile[0][:columns[0]-3] + '..'
            else:
                func = profile[0]
            if len(profile[1]) > columns[1]-2:
                op = profile[1][:columns[1]-3] + '..'
            else:
                op = profile[1]

            summary.append(profile_data['level']*'| ' + column.format(
                func, op, time_spent))

        # footer
        summary.append("*"*total_width)

        for entry in summary:
            self._log(logging.INFO, entry, args)


def setup_logging(loglevel):
    config = {
        'version': 1,
        'formatters': {
            'default': {
                'format': '%(asctime)s %(name)-12s: %(levelname)-8s %(message)s'
            },
            'message': {
                'format': '\033[1;34m%(levelname)-8s\033[0m %(message)s'
            }
        },
        'handlers': {
            'file': {
                'class': 'logging.FileHandler',
                'level': 'DEBUG',
                'formatter': 'default',
                'filename': 'smif.log',
                'mode': 'a',
                'encoding': 'utf-8'
            },
            'stream': {
                'class': 'logging.StreamHandler',
                'formatter': 'message',
                'level': 'DEBUG'
            }
        },
        'root': {
            'handlers': ['file', 'stream'],
            'level': 'DEBUG'
        }
    }

    if loglevel is None:
        config['root']['level'] = logging.WARNING
    elif loglevel == 1:
        config['root']['level'] = logging.INFO
    else:
        config['root']['level'] = logging.DEBUG

    logging.config.dictConfig(config)
    logging.debug('Debug logging enabled.')


logging.Logger.profiling_start = profiling_start
logging.Logger.profiling_stop = profiling_stop
logging.Logger.summary = summary
logging.Logger._profile = OrderedDict()



from typing import Dict, List
from unittest.mock import Mock

from pytest import fixture, raises
from smif.data_layer.store import Store
from smif.decision.decision import DecisionManager, RuleBased
from smif.exception import SmifDataNotFoundError


@fixture(scope='function')
def plan():
    planned_interventions = [
        {'name': 'small_pumping_station_oxford', 'build_year': 2010},
        {'name': 'small_pumping_station_abingdon', 'build_year': 2015},
        {'name': 'large_pumping_station_oxford', 'build_year': 2020}
    ]

    return planned_interventions


@fixture(scope='function')
def get_strategies():
    strategies = [{'strategy': 'pre-specified-planning',
                   'description': 'build_nuclear',
                   'interventions': [
                       {'name': 'nuclear_large', 'build_year': 2012},
                       {'name': 'carrington_retire', 'build_year': 2011}]
                   }]

    return strategies


@fixture(scope='function')
def get_register():
    lifetime = {'technical_lifetime': {'value': 99}}
    register = {'nuclear_large': lifetime,
                'carrington_retire': lifetime,
                'small_pumping_station_oxford': lifetime,
                'small_pumping_station_abingdon': lifetime,
                'large_pumping_station_oxford': lifetime
                }
    return register


class TestRuleBasedProperties:

    def test_timesteps(self):

        timesteps = [2010, 2015, 2020]
        dm = RuleBased(timesteps, Mock())
        assert dm.current_timestep == 2010
        assert dm.next_timestep == 2015

    def test_timesteps_end(self):
        timesteps = [2010, 2015, 2020]
        dm = RuleBased(timesteps, Mock())
        dm.current_timestep = 2020
        assert dm.next_timestep is None

    def test_timesteps_begin(self):
        timesteps = [2010, 2015, 2020]
        dm = RuleBased(timesteps, Mock())
        dm.current_timestep = 2010
        assert dm.previous_timestep is None

    def test_timesteps_first_last(self):

        timesteps = [2010, 2015, 2020]
        dm = RuleBased(timesteps, Mock())
        assert dm.first_timestep == 2010
        assert dm.last_timestep == 2020

    def test_interventions(self):

        all_interventions = {'test_intervention': {'name': 'test_intervention'}}

        timesteps = [2010, 2015, 2020]
        dm = RuleBased(timesteps, all_interventions)
        assert dm.available_interventions([]) == ['test_intervention']

    def test_interventions_planned(self):

        all_interventions = {'test_intervention': {'name': 'test_intervention'},
                             'planned_intervention': {'name': 'planned_intervention'}}

        timesteps = [2010, 2015, 2020]
        dm = RuleBased(timesteps, all_interventions)
        actual = dm.available_interventions([{'name': 'planned_intervention'}])
        expected = ['test_intervention']
        assert actual == expected

    def test_get_intervention(self):

        interventions = {'a': {'name': 'a'},
                         'b': {'name': 'b'},
                         'c': {'name': 'c'}}

        timesteps = [2010, 2015, 2020]
        dm = RuleBased(timesteps, interventions)
        assert dm.get_intervention('a') == interventions['a']
        with raises(SmifDataNotFoundError) as ex:
            dm.get_intervention('z')
        msg = "Intervention 'z' is not found in the list of available interventions"
        assert msg in str(ex.value)


class TestRuleBasedIterationTimestepAccounting:
    """Test that the iteration and timestep accounting methods properly follow
    the path through the decision iterations

    2010 - 0, 1
    2015 - 2, 3
    """

    @fixture(scope='function')
    def dm(self):
        timesteps = [2010, 2015, 2020]
        dm = RuleBased(timesteps, Mock())
        return dm

    def test_first_iteration_base_year(self, dm):

        dm.current_timestep = 2010
        dm.current_iteration = 1
        dm._max_iteration_by_timestep[2010] = 1
        assert dm.get_previous_iteration_timestep() is None

    def test_second_iteration_base_year(self, dm):

        dm.current_timestep = 2010
        dm.current_iteration = 2
        dm._max_iteration_by_timestep[2010] = 2
        assert dm.get_previous_iteration_timestep() == (2010, 1)

    def test_second_iteration_next_year(self, dm):

        dm.current_timestep = 2015
        dm.current_iteration = 3
        dm._max_iteration_by_timestep[2010] = 2
        dm._max_iteration_by_timestep[2015] = 3
        assert dm.get_previous_iteration_timestep() == (2010, 2)

    def test_third_iteration_next_year(self, dm):

        dm.current_timestep = 2015
        dm.current_iteration = 4
        dm._max_iteration_by_timestep[2010] = 2
        dm._max_iteration_by_timestep[2015] = 4
        assert dm.get_previous_iteration_timestep() == (2015, 3)


class TestRuleBased:

    def test_initialisation(self):

        timesteps = [2010, 2015, 2020]
        dm = RuleBased(timesteps, Mock())
        assert dm.timesteps == timesteps
        assert dm.satisfied is False
        assert dm.current_timestep == 2010
        assert dm.current_iteration == 0

    def test_generator(self):

        timesteps = [2010, 2015, 2020]
        dm = RuleBased(timesteps, Mock())

        actual = next(dm)
        assert actual == {
            'decision_iterations': [1],
            'timesteps': [2010]
        }

        dm.satisfied = True
        actual = next(dm)
        assert actual == {
            'decision_iterations': [2],
            'timesteps': [2015],
            'decision_links': {
                2: 1
            }
        }
        assert dm.satisfied is False

        actual = next(dm)
        assert actual == {
            'decision_iterations': [3],
            'timesteps': [2015],
            'decision_links': {
                3: 1
            }
        }
        assert dm.satisfied is False

        dm.satisfied = True
        dm.current_timestep = 2020
        actual = next(dm)
        assert actual is None


class TestDecisionManager():

    @fixture(scope='function')
    def decision_manager(self, empty_store) -> DecisionManager:
        empty_store.write_model_run({'name': 'test', 'sos_model': 'test_sos_model'})
        empty_store.write_sos_model({'name': 'test_sos_model', 'sector_models': []})
        empty_store.write_strategies('test', [])
        sos_model = Mock()
        sos_model.name = 'test_sos_model'
        sos_model.sector_models = []

        df = DecisionManager(empty_store, [2010, 2015], 'test', sos_model)
        return df

    def test_decision_manager_init(self,  decision_manager: DecisionManager):
        df = decision_manager
        dm = df.decision_loop()
        bundle = next(dm)
        assert bundle == {
            'decision_iterations': [0],
            'timesteps': [2010, 2015]
        }
        with raises(StopIteration):
            next(dm)

    def test_available_interventions(self, decision_manager: DecisionManager):
        df = decision_manager
        df._register = {'a': {'name': 'a'},
                        'b': {'name': 'b'},
                        'c': {'name': 'c'}}

        assert df.available_interventions == df._register

        df.planned_interventions = {(2010, 'a'), (2010, 'b')}

        expected = {'c': {'name': 'c'}}

        assert df.available_interventions == expected

    def test_get_intervention(self,  decision_manager: DecisionManager):
        df = decision_manager
        df._register = {'a': {'name': 'a'},
                        'b': {'name': 'b'},
                        'c': {'name': 'c'}}

        assert df.get_intervention('a') == {'name': 'a'}

        with raises(SmifDataNotFoundError):
            df.get_intervention('z')

    def test_buildable(self, decision_manager):

        decision_manager._timesteps = [2010, 2015]
        assert decision_manager.buildable(2010, 2010) is True
        assert decision_manager.buildable(2011, 2010) is True

    def test_historical_intervention_buildable(self, decision_manager):
        decision_manager._timesteps = [2020, 2030]
        assert decision_manager.buildable(1980, 2020) is True
        assert decision_manager.buildable(1990, 2020) is True

    def test_buildable_raises(self, decision_manager):

        with raises(ValueError):
            decision_manager.buildable(2015, 2014)

    def test_within_lifetime(self, decision_manager):

        assert decision_manager.within_lifetime(2010, 2010, 1)

    def test_within_lifetime_does_not_check_start(self, decision_manager):
        """Note that the ``within_lifetime`` method does not check
        that the build year is compatible with timestep
        """
        assert decision_manager.within_lifetime(2011, 2010, 1)

    def test_negative_lifetime_raises(self, decision_manager):
        with raises(ValueError):
            decision_manager.within_lifetime(2010, 2010, -1)


class TestDecisionManagerDecisions:

    @fixture(scope='function')
    def decision_manager(self, empty_store) -> DecisionManager:
        empty_store.write_model_run({'name': 'test', 'sos_model': 'test_sos_model'})
        empty_store.write_sos_model({'name': 'test_sos_model', 'sector_models': []})
        empty_store.write_strategies('test', [])
        sos_model = Mock()
        sos_model.name = 'test_sos_model'
        sos_model.sector_models = []

        interventions = {'test': {'technical_lifetime': {'value': 99}},
                         'planned': {'technical_lifetime': {'value': 99}},
                         'decided': {'technical_lifetime': {'value': 99}}
                         }

        df = DecisionManager(empty_store, [2010, 2015], 'test', sos_model)
        df._register = interventions
        return df

    def test_get_decisions(self, decision_manager: DecisionManager):
        dm = decision_manager

        mock_handle = Mock()
        dm._decision_module = Mock()
        dm._decision_module.get_decision = Mock(
            return_value=[{'name': 'test', 'build_year': 2010}])

        actual = dm._get_decisions(dm._decision_module, mock_handle)
        expected = [(2010, 'test')]
        assert actual == expected

    def test_get_and_save_decisions_dm(self, decision_manager: DecisionManager):
        """Test that the ``get_and_save_decisions`` method updates pre-decision
        state with a new decision and writes it to store
        """
        dm = decision_manager

        dm._decision_module = Mock()
        dm._decision_module.get_decision = Mock(
            return_value=[{'name': 'decided', 'build_year': 2010}])
        dm._decision_module.get_previous_state = Mock(return_value=[])

        dm.get_and_save_decisions(0, 2010)

        actual = dm._store  # type: Store
        expected = [{'name': 'decided', 'build_year': 2010}]  # type: List[Dict]
        assert actual.read_state('test', 2010, decision_iteration=0) == expected

    def test_get_and_save_decisions_prespec(self,
                                            decision_manager: DecisionManager):
        """Test that the ``get_and_save_decisions`` method updates pre-decision
        state with a pre-specified planning and writes it to store
        """
        dm = decision_manager

        dm.planned_interventions = [(2010, 'planned')]

        dm.get_and_save_decisions(0, 2010)

        actual = dm._store  # type: Store
        expected = [{'name': 'planned', 'build_year': 2010}]  # type: List[Dict]
        assert actual.read_state('test', 2010, decision_iteration=0) == expected

    def test_pre_spec_and_decision_module(self,
                                          decision_manager: DecisionManager):
        dm = decision_manager

        dm._decision_module = Mock()
        dm._decision_module.get_decision = Mock(
            return_value=[{'name': 'decided', 'build_year': 2010}])
        dm._decision_module.get_previous_state = Mock(return_value=[])

        dm.planned_interventions = [(2010, 'planned')]

        dm.get_and_save_decisions(0, 2010)

        actual = dm._store.read_state('test', 2010, decision_iteration=0)  # type: List[Dict]

        expected = set([('decided', 2010), ('planned', 2010)])
        assert set([(x['name'], x['build_year']) for x in actual]) == expected



# -*- coding: utf-8 -*-
"""A command line interface to the system of systems framework

This command line interface implements a number of methods.

- `setup` creates an example project with the recommended folder structure
- `run` performs a simulation of an individual sector model, or the whole system
        of systems model
- `validate` performs a validation check of the configuration file
- `app` runs the graphical user interface, opening in a web browser

Folder structure
----------------

When configuring a project for the CLI, the folder structure below should be
used.  In this example, there is one system-of-systems model, combining a water
supply and an energy demand model::

    /config
        project.yaml
        /sector_models
            energy_demand.yml
            water_supply.yml
        /sos_models
            energy_water.yml
        /model_runs
            run_to_2050.yml
            short_test_run.yml
            ...
    /data
        /initial_conditions
            reservoirs.yml
        /interval_definitions
            annual_intervals.csv
        /interventions
            water_supply.yml
        /narratives
            high_tech_dsm.yml
        /region_definitions
            /oxfordshire
                regions.geojson
            /uk_nations_shp
                regions.shp
        /scenarios
            population.csv
            raininess.csv
        /water_supply
            /initial_system
    /models
        energy_demand.py
        water_supply.py
    /planning
        expected_to_2020.yaml
        national_infrastructure_pipeline.yml

The sector model implementations can be installed independently of the model run
configuration. The paths to python wrapper classes (implementing SectorModel)
should be specified in each ``sector_model/*.yml`` configuration.

The project.yaml file specifies the metadata shared by all elements of the
project; ``sos_models`` specify the combinations of ``sector_models`` and
``scenarios`` while individual ``model_runs`` specify the scenario, strategy
and narrative combinations to be used in each run of the models.

"""
from __future__ import print_function

import glob
import logging
import os
import sys
from argparse import ArgumentParser

import pkg_resources

import pandas
import smif
import smif.cli.log
from smif.controller import copy_project_folder, execute_model_run
from smif.controller.run import DAFNIRunScheduler, SubProcessRunScheduler
from smif.data_layer import Store
from smif.data_layer.file import (CSVDataStore, FileMetadataStore,
                                  ParquetDataStore, YamlConfigStore)
from smif.http_api import create_app

try:
    import _thread
except ImportError:
    import thread as _thread

try:
    import win32api

    USE_WIN32 = True
except ImportError:
    USE_WIN32 = False

__author__ = "Will Usher, Tom Russell"
__copyright__ = "Will Usher, Tom Russell"
__license__ = "mit"


def list_model_runs(args):
    """List the model runs defined in the config, optionally indicating whether complete
    results exist.
    """
    store = _get_store(args)
    model_run_configs = store.read_model_runs()

    if args.complete:
        print('Model runs with an asterisk (*) have complete results available\n')

    for run in model_run_configs:
        run_name = run['name']

        if args.complete:
            expected_results = store.canonical_expected_results(run_name)
            available_results = store.canonical_available_results(run_name)

            complete = ' *' if expected_results == available_results else ''

            print('{}{}'.format(run_name, complete))
        else:
            print(run_name)


def list_available_results(args):
    """List the available results for a specified model run.
    """

    store = _get_store(args)
    expected = store.canonical_expected_results(args.model_run)
    available = store.available_results(args.model_run)

    # Print run and sos model
    run = store.read_model_run(args.model_run)
    print('\nmodel run: {}'.format(args.model_run))
    print('{}- sos model: {}'.format(' ' * 2, run['sos_model']))

    # List of expected sector models
    sec_models = sorted({sec for _t, _d, sec, _out in expected})

    for sec_model in sec_models:
        print('{}- sector model: {}'.format(' ' * 4, sec_model))

        # List expected outputs for this sector model
        outputs = sorted({out for _t, _d, sec, out in expected if sec == sec_model})

        for output in outputs:
            print('{}- output: {}'.format(' ' * 6, output))

            # List available decisions for this sector model and output
            decs = sorted({d for _t, d, sec, out in available if
                           sec == sec_model and out == output})

            if len(decs) == 0:
                print('{}- no results'.format(' ' * 8))

            for dec in decs:
                base_str = '{}- decision {}:'.format(' ' * 8, dec)

                # List available time steps for this decision, sector model and output
                ts = sorted({t for t, d, sec, out in available if
                             d == dec and sec == sec_model and out == output})
                assert (len(
                    ts) > 0), "If a decision is available, so is at least one time step"

                res_str = ', '.join([str(t) for t in ts])
                print('{} {}'.format(base_str, res_str))


def list_missing_results(args):
    """List the missing results for a specified model run.
    """
    store = _get_store(args)
    expected = store.canonical_expected_results(args.model_run)
    missing = store.canonical_missing_results(args.model_run)

    # Print run and sos model
    run = store.read_model_run(args.model_run)
    print('\nmodel run: {}'.format(args.model_run))
    print('{}- sos model: {}'.format(' ' * 2, run['sos_model']))

    # List of expected sector models
    sec_models = sorted({sec for _t, _d, sec, _out in expected})

    for sec_model in sec_models:
        print('{}- sector model: {}'.format(' ' * 4, sec_model))

        # List expected outputs for this sector model
        outputs = sorted({out for _t, _d, sec, out in expected if sec == sec_model})

        for output in outputs:
            print('{}- output: {}'.format(' ' * 6, output))

            # List missing time steps for this sector model and output
            ts = sorted({t for t, d, sec, out in missing if
                         sec == sec_model and out == output})

            if len(ts) == 0:
                print('{}- no missing results'.format(' ' * 8))
            else:
                base_str = '{}- results missing for:'.format(' ' * 8)
                res_str = ', '.join([str(t) for t in ts])
                print('{} {}'.format(base_str, res_str))


def prepare_convert(args):
    src_store = _get_store(args)
    if isinstance(src_store.data_store, CSVDataStore):
        tgt_store = Store(
            config_store=YamlConfigStore(args.directory),
            metadata_store=FileMetadataStore(args.directory),
            data_store=ParquetDataStore(args.directory),
            model_base_folder=(args.directory)
        )
    else:
        tgt_store = Store(
            config_store=YamlConfigStore(args.directory),
            metadata_store=FileMetadataStore(args.directory),
            data_store=CSVDataStore(args.directory),
            model_base_folder=(args.directory)
        )
    # Read model run
    model_run = src_store.read_model_run(args.model_run)
    # Read sos model for model run
    sos_model = src_store.read_sos_model(model_run['sos_model'])
    # Now let us convert data
    # Convert strategies interventions for model run
    src_store.convert_strategies_data(model_run['name'], tgt_store, args.noclobber)
    # Convert scenario data for model run
    src_store.convert_scenario_data(model_run['name'], tgt_store)
    # Convert narrative data for sos model
    src_store.convert_narrative_data(sos_model['name'], tgt_store, args.noclobber)
    # Convert initial conditions, default parameter and interventions data
    # for sector models in sos model
    for sector_model_name in sos_model['sector_models']:
        src_store.convert_model_parameter_default_data(
            sector_model_name, tgt_store, args.noclobber)
        src_store.convert_interventions_data(sector_model_name, tgt_store, args.noclobber)
        src_store.convert_initial_conditions_data(sector_model_name, tgt_store, args.noclobber)


def csv2parquet(args):
    """Convert CSV to Parquet - assuming the CSV can be parsed as a dataframe
    """
    path = args.path
    if ".csv" in path:
        files = [path]
    else:
        files = glob.glob(os.path.join(path, '**', '*.csv'), recursive=True)

    for csv_path in files:
        parquet_path = csv_path.replace(".csv", ".parquet")
        if args.noclobber and os.path.exists(parquet_path):
            print("Skipping", csv_path)
        else:
            print("Converting", csv_path, flush=True)
            try:
                dataframe = pandas.read_csv(csv_path)
                dataframe.to_parquet(parquet_path, engine='pyarrow', compression='gzip')
            except UnicodeDecodeError:
                # guess that cp1252 is next most common encoding we'll come across
                dataframe = pandas.read_csv(csv_path, encoding='cp1252')
                dataframe.to_parquet(parquet_path, engine='pyarrow', compression='gzip')
            except pandas.errors.ParserError as ex:
                # nothing we can do with ParserError - usually a data problem
                print(ex)
                continue


def prepare_scenario(args):
    """Update scenario configuration file to include multiple scenario variants.

    The initial scenario configuration file is overwritten.
    """
    # Read template scenario using the Store class
    store = _get_store(args)
    list_of_variants = range(args.variants_range[0], args.variants_range[1] + 1)

    store.prepare_scenario(args.scenario_name, list_of_variants)


def prepare_model_runs(args):
    """Generate multiple model runs according to a model run file referencing a scenario
    with multiple variants.
    """
    # Read model run and scenario using the Store class
    store = _get_store(args)
    nb_variants = len(store.read_scenario_variants(args.scenario_name))
    # Define default lower and upper of variant range
    var_start = 0
    var_end = nb_variants

    # Check if optional cli arguments specify range of variants
    # They are compared to None because they can be 0
    if args.start is not None:
        var_start = args.start
        if var_start < 0:
            raise ValueError('Lower bound of variant range must be >=0')
        if var_start > nb_variants:
            raise ValueError("Lower bound of variant range greater"
                             " than number of variants")
    if args.end is not None:
        var_end = args.end
        if var_end < 0:
            raise ValueError('Upper bound of variant range must be >=0')
        if var_end > nb_variants - 1:
            raise ValueError("Upper bound of variant range cannot be greater"
                             " than {:d}".format(nb_variants - 1))
        if var_end < var_start:
            raise ValueError("Upper bound of variant range must be >= lower"
                             " bound of variant range")

    store.prepare_model_runs(args.model_run_name, args.scenario_name,
                             var_start, var_end)


def run_model_runs(args):
    """Run the model runs as requested. Check if results exist and asks
    user for permission to overwrite

    Parameters
    ----------
    args
    """
    logger = logging.getLogger(__name__)
    try:
        logger.profiling_start('run_model_runs', '{:s}, {:s}, {:s}'.format(
            args.modelrun, args.interface, args.directory))
    except AttributeError:
        logger.info('START run_model_runs', '{:s}, {:s}, {:s}'.format(
            args.modelrun, args.interface, args.directory))

    if args.batchfile:
        with open(args.modelrun, 'r') as f:
            model_run_ids = f.read().splitlines()
    else:
        model_run_ids = [args.modelrun]

    store = _get_store(args)
    execute_model_run(model_run_ids, store, args.warm)
    try:
        logger.profiling_stop('run_model_runs', '{:s}, {:s}, {:s}'.format(
            args.modelrun, args.interface, args.directory))
        logger.summary()
    except AttributeError:
        logger.info('STOP run_model_runs', '{:s}, {:s}, {:s}'.format(
            args.modelrun, args.interface, args.directory))


def _get_store(args):
    """Contruct store as configured by arguments
    """
    if args.interface == 'local_csv':
        store = Store(
            config_store=YamlConfigStore(args.directory),
            metadata_store=FileMetadataStore(args.directory),
            data_store=CSVDataStore(args.directory),
            model_base_folder=args.directory
        )
    elif args.interface == 'local_binary':
        store = Store(
            config_store=YamlConfigStore(args.directory),
            metadata_store=FileMetadataStore(args.directory),
            data_store=ParquetDataStore(args.directory),
            model_base_folder=args.directory
        )
    else:
        raise ValueError("Store interface type {} not recognised.".format(args.interface))
    return store


def _run_server(args):
    app_folder = pkg_resources.resource_filename('smif', 'app/dist')
    if args.scheduler == 'dafni' and args.interface != 'local_csv':
        msg = "Scheduler implementation {0}, is not valid when combined with {1}."
        raise ValueError(msg.format(args.scheduler, args.interface))

    if args.scheduler == 'default':
        model_scheduler = SubProcessRunScheduler()
    elif args.scheduler == 'dafni':
        model_scheduler = DAFNIRunScheduler(args.username, args.password)
    else:
        raise ValueError("Scheduler implentation {} not recognised.".format(args.scheduler))

    app = create_app(
        static_folder=app_folder,
        template_folder=app_folder,
        data_interface=_get_store(args),
        scheduler=model_scheduler
    )

    print("    Opening smif app\n")
    print("    Copy/paste this URL into your web browser to connect:")
    print("        http://localhost:" + str(args.port) + "\n")
    # add flush to ensure that text is printed before server thread starts
    print("    Close your browser then type Control-C here to quit.", flush=True)
    app.run(host='0.0.0.0', port=args.port, threaded=True)


def run_app(args):
    """Run the client/server application

    Parameters
    ----------
    args
    """
    # avoid one of two error messages from 'forrtl error(200)' when running
    # on windows cmd - seems related to scipy's underlying Fortran
    os.environ['FOR_DISABLE_CONSOLE_CTRL_HANDLER'] = 'T'

    if USE_WIN32:
        # Set handler for CTRL-C. Necessary to avoid `forrtl: error (200):
        # program aborting...` crash on CTRL-C if we're runnging from Windows
        # cmd.exe
        def handler(dw_ctrl_type, hook_sigint=_thread.interrupt_main):
            """Handler for CTRL-C interrupt
            """
            if dw_ctrl_type == 0:  # CTRL-C
                hook_sigint()
                return 1  # don't chain to the next handler
            return 0  # chain to the next handler

        win32api.SetConsoleCtrlHandler(handler, 1)

    # Create backend server process
    _run_server(args)


def setup_project_folder(args):
    """Setup a sample project
    """
    copy_project_folder(args.directory)


def parse_arguments():
    """Parse command line arguments

    Returns
    =======
    :class:`argparse.ArgumentParser`

    """
    parser = ArgumentParser(description='Command line tools for smif')
    parser.add_argument('-V', '--version',
                        action='version',
                        version="smif " + smif.__version__,
                        help='show the current version of smif')

    parent_parser = ArgumentParser(add_help=False)
    parent_parser.add_argument('-v', '--verbose',
                               action='count',
                               help='show messages: -v to see messages reporting on ' +
                                    'progress, -vv to see debug messages.')
    parent_parser.add_argument('-i', '--interface',
                               default='local_csv',
                               choices=['local_csv', 'local_binary'],
                               help="Select the data interface (default: %(default)s)")
    parent_parser.add_argument('-d', '--directory',
                               default='.',
                               help="Path to the project directory")

    subparsers = parser.add_subparsers(help='available commands')

    # SETUP
    parser_setup = subparsers.add_parser(
        'setup', help='Setup the project folder', parents=[parent_parser])
    parser_setup.set_defaults(func=setup_project_folder)

    # LIST
    parser_list = subparsers.add_parser(
        'list', help='List available model runs', parents=[parent_parser])
    parser_list.set_defaults(func=list_model_runs)
    parser_list.add_argument('-c', '--complete',
                             help="Show which model runs have complete results",
                             action='store_true')

    # RESULTS
    parser_available_results = subparsers.add_parser(
        'available_results', help='List available results', parents=[parent_parser])
    parser_available_results.set_defaults(func=list_available_results)
    parser_available_results.add_argument(
        'model_run',
        help="Name of the model run to list available results"
    )

    parser_missing_results = subparsers.add_parser(
        'missing_results', help='List missing results', parents=[parent_parser])
    parser_missing_results.set_defaults(func=list_missing_results)
    parser_missing_results.add_argument(
        'model_run',
        help="Name of the model run to list missing results"
    )

    # PREPARE
    parser_convert = subparsers.add_parser(
        'prepare-convert', help='Convert data from one format to another',
        parents=[parent_parser])
    parser_convert.set_defaults(func=prepare_convert)
    parser_convert.add_argument(
        'model_run', help='Name of the model run')
    parser_convert.add_argument(
        '-nc', '--noclobber',
        help='Do not convert existing data files', action='store_true')

    parser_prepare_scenario = subparsers.add_parser(
        'prepare-scenario', help='Prepare scenario configuration file with multiple variants',
        parents=[parent_parser])
    parser_prepare_scenario.set_defaults(func=prepare_scenario)
    parser_prepare_scenario.add_argument(
        'scenario_name', help='Name of the scenario')
    parser_prepare_scenario.add_argument(
        'variants_range', nargs=2, type=int,
        help='Two integers delimiting the range of variants')

    parser_prepare_model_runs = subparsers.add_parser(
        'prepare-run', help='Prepare model runs based on scenario variants',
        parents=[parent_parser])
    parser_prepare_model_runs.set_defaults(func=prepare_model_runs)
    parser_prepare_model_runs.add_argument(
        'scenario_name', help='Name of the scenario')
    parser_prepare_model_runs.add_argument(
        'model_run_name', help='Name of the template model run')
    parser_prepare_model_runs.add_argument(
        '-s', '--start', type=int, help='Lower bound of the range of variants')
    parser_prepare_model_runs.add_argument(
        '-e', '--end', type=int, help='Upper bound of the range of variants')

    # CONVERT
    parser_convert_format = subparsers.add_parser(
        'csv2parquet', help='Convert CSV to Parquet. Pass a filename or a directory to ' +
        'search recurisvely', parents=[parent_parser])
    parser_convert_format.set_defaults(func=csv2parquet)
    parser_convert_format.add_argument(
        'path', help='Path to file')
    parser_convert_format.add_argument(
        '-nc', '--noclobber',
        help='Skip converting data files which already exist as parquet', action='store_true')

    # APP
    parser_app = subparsers.add_parser(
        'app', help='Open smif app', parents=[parent_parser])
    parser_app.set_defaults(func=run_app)
    parser_app.add_argument('-p', '--port',
                            type=int,
                            default=5000,
                            help="The port over which to serve the app")
    parser_app.add_argument('-s', '--scheduler',
                            default='default',
                            choices=['default', 'dafni'],
                            help="The module scheduling implementation to use")
    parser_app.add_argument('-u', '--username',
                            help="The username for logging in to the dafni JobSubmissionAPI, \
                                  only needed with the dafni job scheduler")
    parser_app.add_argument('-pw', '--password',
                            help="The password for logging in to the dafni JobSubmissionAPI, \
                                  only needed with the dafni job scheduler")

    # RUN
    parser_run = subparsers.add_parser(
        'run', help='Run a model', parents=[parent_parser])
    parser_run.set_defaults(func=run_model_runs)
    parser_run.add_argument('-w', '--warm',
                            action='store_true',
                            help="Use intermediate results from the last modelrun \
                                  and continue from where it had left")
    parser_run.add_argument('-b', '--batchfile',
                            action='store_true',
                            help="Use a batchfile instead of a modelrun name (a \
                                  list of modelrun names)")
    parser_run.add_argument('modelrun',
                            help="Name of the model run to run")

    return parser


def confirm(prompt=None, response=False):
    """Prompts for a yes or no response from the user

    Arguments
    ---------
    prompt : str, default=None
    response : bool, default=False

    Returns
    -------
    bool
        True for yes and False for no.

    Notes
    -----

    `response` should be set to the default value assumed by the caller when
    user simply types ENTER.

    Examples
    --------

    >>> confirm(prompt='Create Directory?', response=True)
    Create Directory? [y]|n:
    True
    >>> confirm(prompt='Create Directory?', response=False)
    Create Directory? [n]|y:
    False
    >>> confirm(prompt='Create Directory?', response=False)
    Create Directory? [n]|y: y
    True

    """

    if prompt is None:
        prompt = 'Confirm'

    if response:
        prompt = '{} [{}]|{}: '.format(prompt, 'y', 'n')
    else:
        prompt = '{} [{}]|{}: '.format(prompt, 'n', 'y')

    while True:
        ans = input(prompt)
        if not ans:
            return response
        if ans not in ['y', 'Y', 'n', 'N']:
            print('please enter y or n.')
            continue
        if ans in ['y', 'Y']:
            return True
        if ans in ['n', 'N']:
            return False


def main(arguments=None):
    """Parse args and run
    """
    parser = parse_arguments()
    args = parser.parse_args(arguments)

    try:
        smif.cli.log.setup_logging(args.verbose)
    except AttributeError:
        # verbose is only set on subcommands - so `smif` or `smif -h` would error
        pass

    def exception_handler(exception_type, exception, traceback, debug_hook=sys.excepthook):
        if args.verbose:
            debug_hook(exception_type, exception, traceback)
        else:
            print("{}: {}".format(exception_type.__name__, exception))

    sys.excepthook = exception_handler
    if 'func' in args:
        args.func(args)
    else:
        parser.print_help()



"""Schedulers are used to run models.

The defaults provided allow model runs to be scheduled as subprocesses, or individual models to
be called in series.

Future implementations may interface with common schedulers to queue up models to run in
parallel and/or distributed.

The DAFNIRunScheduler currently only works while connected to the RAL vpn.

Posts all of the model run information to the DAFNI API so that a DAFNI worker can start the
model run.
"""
import json
import os
import time
from collections import defaultdict

import requests
from minio import Minio
from minio.error import ResponseError
from ruamel.yaml import YAML  # type: ignore

if "BACKEND_NISMOD_MINIO_SECRETS_FILE" in os.environ:
    MINIO_CREDENTIALS_FILE = os.environ['BACKEND_NISMOD_MINIO_SECRETS_FILE']
else:
    MINIO_CREDENTIALS_FILE = ""
if "BACKEND_SECRET_KEY" in os.environ:
    SECRET_KEY = os.environ['BACKEND_SECRET_KEY']
else:
    SECRET_KEY = ""
if "BACKEND_ACCESS_KEY" in os.environ:
    ACCESS_KEY = os.environ['BACKEND_ACCESS_KEY']
else:
    ACCESS_KEY = ""
if "BACKEND_JOBSUBMISSION_API" in os.environ:
    JOBSUBMISSION_API_URL = os.environ['BACKEND_JOBSUBMISSION_API']
else:
    JOBSUBMISSION_API_URL = ""
if "BACKEND_MINIO_IP" in os.environ:
    MINIO_IP = os.environ['BACKEND_MINIO_IP']
else:
    MINIO_IP = ""

URL_AUTH = JOBSUBMISSION_API_URL + "auth/obtain_token/"
URL_JOBS = JOBSUBMISSION_API_URL + "nismod-model/jobs"


class DAFNIRunScheduler(object):
    """The scheduler can run instances of smif as a subprocess
    and can provide information whether the modelrun is running,
    is done or has failed.
    """
    def __init__(self, username, password):
        self._status = defaultdict(lambda: 'unstarted')
        self._process = {}
        self._output = defaultdict(str)
        self._err = {}
        self.jobId = 0
        self.lock = False
        self.username = username
        self.password = password
        response = requests.post(
            URL_AUTH,
            json={
                "username": self.username,
                "password": self.password
            },
            allow_redirects=False
        )
        response.raise_for_status()
        token = response.json()['token']
        self.auth_header = json.loads('{ "Authorization": "JWT ' + token + '"}')
        response = requests.get(URL_JOBS, headers=self.auth_header)
        response.raise_for_status()

    def add(self, model_run_name, args):
        """Add a model_run to the Modelrun scheduler.

        Parameters
        ----------
        model_run_name: str
            Name of the modelrun
        args: dict
            Arguments for the command-line interface

        Exception
        ---------
        Exception
            When the modelrun was already started

        Notes
        -----
        DAFNI's queuing mechanism starts model runs in separate
        container. This means that it is possible to run multiple
        modelruns concurrently. This will not cause conflicts.
        """
        if self._status[model_run_name] != 'running':
            self._output[model_run_name] = ''
            self._status[model_run_name] = 'queing'

            yaml_files = self.get_yamls(model_run_name, args)
            model_run_id = model_run_name.replace("_", "-")

            minio_credentials = self.get_dict_from_json(MINIO_CREDENTIALS_FILE)
            minio_client = Minio(
                MINIO_IP,
                access_key=minio_credentials['accessKey'],
                secret_key=minio_credentials['secretKey'],
                secure=False
            )
            bucket_list = minio_client.list_buckets()

            for bucket in bucket_list:
                if bucket.name == model_run_id:
                    for obj in minio_client.list_objects(model_run_id, recursive=True):
                        minio_client.remove_object(model_run_id, obj.object_name)
                    minio_client.remove_bucket(model_run_id)

            minio_client.make_bucket(model_run_id)
            for yml in yaml_files:
                try:
                    local_path = args['directory'] + yml
                    with open(local_path, 'rb') as yml_data:
                        yml_stat = os.stat(local_path)
                        minio_client.put_object(
                            model_run_id, yml[1:], yml_data, yml_stat.st_size)
                except ResponseError as err:
                    print(err)

            response = requests.get(URL_JOBS, headers=self.auth_header)
            response.raise_for_status()

            for job in response.json():
                if job['job']['job_name'] == model_run_id:
                    response = requests.delete(
                        URL_JOBS + "/" + str(job['job']['id']), headers=self.auth_header)
                    response.raise_for_status()

            response = requests.post(
                URL_JOBS,
                json={
                    "job_name": model_run_id,
                    "model_name": model_run_name,
                    "minio_config_id": model_run_id
                },
                headers=self.auth_header
            )
            response.raise_for_status()

    def get_scheduler_type(self):
        return "dafni"

    def get_yamls(self, model_run_name, args):
        yaml_files = []
        yaml_files.append("/config/model_runs/" + model_run_name + ".yml")
        f = open(args['directory'] + yaml_files[0], "r")
        doc = YAML(typ='safe').load(f.read())

        yaml_files.append("/config/sos_models/" + doc['sos_model'] + ".yml")

        sos_f = open(args['directory'] + yaml_files[1])
        sos_doc = YAML(typ='safe').load(sos_f.read())

        for sector_model in sos_doc['sector_models']:
            yaml_files.append("/config/sector_models/" + sector_model + ".yml")

        for scenario in sos_doc['scenarios']:
            yaml_files.append("/config/scenarios/" + scenario + ".yml")

        return yaml_files

    def get_dict_from_json(self, file_path):
        '''
        Given a JSON file, will return the dictionary of values within that file.
        Is developed to wait for a file to exist (for the case of reading vault secrets)
        to avoid the inherent race condition.
        '''
        count = 0
        while True:
            try:
                with open(file_path) as file:
                    data = json.load(file)
                    return data
            except FileNotFoundError:
                count += 1
                if count > 3:
                    raise FileNotFoundError
                else:
                    print('{} does not exist yet. Waiting 5 seconds and trying again.\
                        Have tried {} times'.format(file_path, count))
                    time.sleep(5)

    def kill(self, model_run_name):
        if self._status[model_run_name] == 'running':
            self._status[model_run_name] = 'stopped'

        minio_credentials = self.get_dict_from_json(MINIO_CREDENTIALS_FILE)
        minio_client = Minio(
            MINIO_IP,
            access_key=minio_credentials['accessKey'],
            secret_key=minio_credentials['secretKey'],
            secure=False
        )

        model_run_id = model_run_name.replace("_", "-")
        yaml_files_minio = minio_client.list_objects(model_run_id, recursive=True)
        for d in yaml_files_minio:
            minio_client.remove_object(model_run_id, d.object_name)

        minio_client.remove_bucket(model_run_id)

        response = requests.get(URL_JOBS, headers=self.auth_header)
        response.raise_for_status()

        for job in response.json():
            if job['job']['job_name'] == model_run_id:
                requests.delete(
                    URL_JOBS + "/" + str(job['job']['id']), headers=self.auth_header)

    def get_status(self, model_run_name):
        response = requests.get(URL_JOBS, headers=self.auth_header)
        response.raise_for_status()
        model_run_id = model_run_name.replace("_", "-")
        if len(response.json()) > 0:
            for j in response.json():
                if j['job']['job_name'] == model_run_id:
                    job = j['job']
                    status = job['status']
                    jobStatus = ["unstarted", "unstarted", "running", "done", "failed"]
                    self._status[model_run_name] = jobStatus[status]
                    break
        else:
            self._status[model_run_name] = "unstarted"
        return {
            'status': self._status[model_run_name],
            'output': self._output[model_run_name]
        }



"""Test all DataStore implementations
"""
from copy import deepcopy

import numpy as np
from pytest import fixture, mark, param, raises
from smif.data_layer.data_array import DataArray
from smif.data_layer.database_interface import DbDataStore
from smif.data_layer.file.file_data_store import CSVDataStore, ParquetDataStore
from smif.data_layer.memory_interface import MemoryDataStore
from smif.exception import SmifDataNotFoundError
from smif.metadata import Spec


@fixture(
    params=[
        'memory',
        'file_csv',
        'file_parquet',
        param('database', marks=mark.skip)]
    )
def handler(request, setup_empty_folder_structure):
    if request.param == 'memory':
        handler = MemoryDataStore()
    elif request.param == 'file_csv':
        base_folder = setup_empty_folder_structure
        handler = CSVDataStore(base_folder)
    elif request.param == 'file_parquet':
        base_folder = setup_empty_folder_structure
        handler = ParquetDataStore(base_folder)
    elif request.param == 'database':
        handler = DbDataStore()
        raise NotImplementedError

    return handler


class TestDataArray():
    """Read and write DataArray
    """
    def test_read_write_data_array(self, handler, scenario):
        spec_config = deepcopy(scenario['provides'][0])
        spec_config['dims'] = ['timestep'] + spec_config['dims']
        spec_config['coords']['timestep'] = [{'name': 2010}]
        spec = Spec.from_dict(spec_config)
        data = np.array([[0, 1]], dtype='float')
        da = DataArray(spec, data)
        handler.write_scenario_variant_data('mortality.csv', da)

        spec_config = deepcopy(scenario['provides'][0])
        spec = Spec.from_dict(spec_config)
        data = np.array([0, 1], dtype='float')
        expected = DataArray(spec, data)

        actual = handler.read_scenario_variant_data('mortality.csv', spec, 2010)
        assert actual == expected
        np.testing.assert_array_equal(actual.as_ndarray(), expected.as_ndarray())

    def test_read_write_data_array_all(self, handler, scenario):
        spec = Spec.from_dict(deepcopy(scenario['provides'][0]))

        spec_with_t = scenario['provides'][0]
        spec_with_t['dims'].insert(0, 'timestep')
        spec_with_t['coords']['timestep'] = [2010, 2015]
        spec_with_t = Spec.from_dict(spec_with_t)
        da = DataArray(spec_with_t, np.array([[0, 1], [2, 3]], dtype='float'))

        handler.write_scenario_variant_data('mortality.csv', da)
        actual = handler.read_scenario_variant_data('mortality.csv', spec_with_t)
        expected = np.array([[0, 1], [2, 3]], dtype='float')
        np.testing.assert_array_equal(actual.as_ndarray(), expected)

        da_2010 = handler.read_scenario_variant_data('mortality.csv', spec, 2010)
        expected = np.array([0, 1], dtype='float')
        np.testing.assert_array_equal(da_2010.as_ndarray(), expected)

        da_2015 = handler.read_scenario_variant_data('mortality.csv', spec, 2015)
        expected = np.array([2, 3], dtype='float')
        np.testing.assert_array_equal(da_2015.as_ndarray(), expected)

    def test_read_zero_d_from_timeseries(self, handler):
        """Read a single value
            timestep,param
            2010,0
            2015,1
        """
        # write data for multiple timesteps by adding 'timestep' dim to the spec
        data = np.array([0, 1], dtype=float)
        write_spec = Spec(
            name='param',
            dims=['timestep'],
            coords={'timestep': [2010, 2015]},
            dtype='float'
        )
        da = DataArray(write_spec, data)
        handler.write_scenario_variant_data('param', da)

        read_spec = Spec(
            name='param',
            dims=[],
            coords={},
            dtype='float'
        )
        actual = handler.read_scenario_variant_data('param', read_spec, 2010).data
        assert actual == np.array(0.0)
        actual = handler.read_scenario_variant_data('param', read_spec, 2015).data
        assert actual == np.array(1.0)

    def test_read_data_array_missing_timestep(self, handler, scenario):
        data = np.array([[0, 1]], dtype=float)
        spec_config = deepcopy(scenario['provides'][0])
        spec_config['dims'] = ['timestep'] + spec_config['dims']
        spec_config['coords']['timestep'] = [{'name': 2010}]
        spec = Spec.from_dict(spec_config)

        da = DataArray(spec, data)

        handler.write_scenario_variant_data('mortality.csv', da)
        msg = "not found for timestep 2011"
        with raises(SmifDataNotFoundError) as ex:
            handler.read_scenario_variant_data('mortality.csv', spec, 2011)
        assert msg in str(ex.value)

    def test_string_data(self, handler):
        spec = Spec(
            name='string_data',
            dims=['timestep', 'zones'],
            coords={'timestep': [2010], 'zones': ['a', 'b', 'c']},
            dtype='object'
        )
        data = np.array([['alpha', 'beta', 'γάμμα']], dtype='object')
        expected = DataArray(spec, data)

        handler.write_scenario_variant_data('key', expected)
        actual = handler.read_scenario_variant_data('key', spec)
        assert actual == expected


class TestInitialConditions():
    """Read and write initial conditions
    """
    def test_read_write_initial_conditions(self, handler, initial_conditions):

        expected = initial_conditions

        handler.write_initial_conditions('initial_conditions.csv', initial_conditions)
        actual = handler.read_initial_conditions(['initial_conditions.csv'])

        assert actual == expected


class TestInterventions():
    """Read and write interventions
    """
    def test_read_write_interventions(self, handler, interventions):

        expected = interventions
        handler.write_interventions('my_intervention.csv', interventions)
        actual = handler.read_interventions(['my_intervention.csv'])

        assert actual == expected


class TestState():
    """Read and write state
    """
    def test_read_write_state(self, handler, state):
        expected = state
        modelrun_name = 'test_modelrun'
        timestep = 2020
        decision_iteration = None

        handler.write_state(expected, modelrun_name, timestep, decision_iteration)
        actual = handler.read_state(modelrun_name, timestep, decision_iteration)
        assert actual == expected

    def test_read_write_empty_state(self, handler):
        expected = []
        modelrun_name = 'test_modelrun'
        timestep = 2020
        decision_iteration = None

        handler.write_state(expected, modelrun_name, timestep, decision_iteration)
        actual = handler.read_state(modelrun_name, timestep, decision_iteration)
        assert actual == expected


class TestCoefficients():
    """Read/write conversion coefficients
    """
    def test_read_write_coefficients(self, handler):

        with raises(SmifDataNotFoundError):
            handler.read_coefficients('from_dim_name', 'to_dim_name')

        expected = np.array([[2]])
        handler.write_coefficients('from_dim_name', 'to_dim_name', expected)

        actual = handler.read_coefficients('from_dim_name', 'to_dim_name')
        np.testing.assert_equal(actual, expected)


class TestResults():
    """Read/write results and prepare warm start
    """
    def test_read_write_results(self, handler, sample_results):
        output_spec = sample_results.spec
        modelrun_name = 'test_modelrun'
        model_name = 'energy'
        timestep = 2010

        handler.write_results(sample_results, modelrun_name, model_name, timestep)
        results_out = handler.read_results(modelrun_name, model_name, output_spec, timestep)

        assert results_out == sample_results

    def test_available_results(self, handler, sample_results):
        """Available results should return an empty list if none are available
        develop
        """
        assert handler.available_results('test_modelrun') == []
        handler.write_results(sample_results, 'test_modelrun', 'energy', 2010, 0)
        handler.write_results(sample_results, 'test_modelrun', 'energy', 2015, 0)
        handler.write_results(sample_results, 'test_modelrun', 'energy', 2015, 1)

        # keys should be (timestep, decision_iteration, model_name, output_name)
        assert sorted(handler.available_results('test_modelrun')) == [
            (2010, 0, 'energy', sample_results.spec.name),
            (2015, 0, 'energy', sample_results.spec.name),
            (2015, 1, 'energy', sample_results.spec.name)
        ]

        # delete one
        handler.delete_results('test_modelrun', 'energy', sample_results.spec.name, 2010, 0)
        assert sorted(handler.available_results('test_modelrun')) == [
            (2015, 0, 'energy', sample_results.spec.name),
            (2015, 1, 'energy', sample_results.spec.name)
        ]

        # clear all (no error if re-deleting)
        handler.delete_results('test_modelrun', 'energy', sample_results.spec.name, 2010, 0)
        handler.delete_results('test_modelrun', 'energy', sample_results.spec.name, 2015, 0)
        handler.delete_results('test_modelrun', 'energy', sample_results.spec.name, 2015, 1)
        assert not handler.available_results('test_modelrun')

    def test_read_results_raises(self, handler, sample_results):
        modelrun_name = 'test_modelrun'
        model_name = 'energy'
        timestep = 2010
        output_spec = sample_results.spec

        handler.write_results(sample_results, modelrun_name, model_name, timestep=timestep)

        with raises(SmifDataNotFoundError):
            handler.read_results(modelrun_name, model_name, output_spec, 2020)


class TestDataExists():
    """Check that model run data exists
    """
    def test_strategy_data_exists(self, handler, strategies):
        for strategy in strategies:
            if strategy['type'] == 'pre-specified-planning':
                new_strategy = {'filename': 'strategy'}
                assert not handler.strategy_data_exists(new_strategy)
                handler.write_strategy_interventions(new_strategy, strategy['interventions'])

                assert handler.strategy_data_exists(new_strategy)

    def test_scenario_variant_data_exists(self, handler, sample_scenario_data):
        assert not handler.scenario_variant_data_exists('scenario_variant_data')

        key = next(iter(sample_scenario_data))
        scenario_name, variant_name, variable = key
        scenario_variant_data = sample_scenario_data[key]
        handler.write_scenario_variant_data('scenario_variant_data', scenario_variant_data)

        assert handler.scenario_variant_data_exists('scenario_variant_data')

    def test_narrative_variant_data_exists(self, handler, sample_narrative_data):
        assert not handler.narrative_variant_data_exists('narrative_variant_data')
        # pick out single sample
        key = (
            'energy',
            'technology',
            'high_tech_dsm',
            'smart_meter_savings'
        )
        sos_model_name, narrative_name, variant_name, param_name = key
        narrative_variant_data = sample_narrative_data[key]
        handler.write_narrative_variant_data('narrative_variant_data', narrative_variant_data)

        assert handler.narrative_variant_data_exists('narrative_variant_data')

    def test_model_parameter_default_data_exists(self, handler, get_multidimensional_param):
        assert not handler.model_parameter_default_data_exists('parameter_default')
        param_data = get_multidimensional_param
        handler.write_model_parameter_default('parameter_default', param_data)

        assert handler.model_parameter_default_data_exists('parameter_default')

    def test_interventions_data_exists(self, handler, interventions):
        assert not handler.interventions_data_exists('interventions')
        handler.write_interventions('interventions', interventions)

        assert handler.interventions_data_exists('interventions')

    def test_initial_conditions_data_exists(self, handler, initial_conditions):
        assert not handler.initial_conditions_data_exists('initial_conditions')
        handler.write_initial_conditions('initial_conditions', initial_conditions)

        assert handler.initial_conditions_data_exists('initial_conditions')
